/*******************************************************************************
 *	Test2.c			CHASE_DA test code
 *				T.Barnaby,	BEAM Ltd,	2009-04-21
 *******************************************************************************
 *
 * This program generates a sawtooth waveform on all channels.
 *
 * On a DA8150 the waveform frequency is: 2.34MHz (150/64).
 * On a DA11000 the waveform frequency is: 15.625MHz (1G/64).
 *
 * Copyright (c) 2009 BEAM Ltd. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>
#include <errno.h>
#include <math.h>
#include <chase_da_lib.h>

int main(){
	int		i;
	int		cardNum = 1;
	int		nPoints = 1*64;
	int		r;
	ChaseDaSegment	segs[1];
	uint16_t	data0[nPoints];
	
	i = chase_da_CountCards();
	printf("There are %d Chase_da cards in the system\n", i);
	
	printf("Open Card: %d\n", cardNum);
	if(r = chase_da_Open(cardNum)){
		fprintf(stderr, "Error unable to open CHASE_DA: %d\n", r);
		return 1;
	}

	if(chase_da_BoardType(cardNum) == CHASE_DA8150){
		printf("Set Clock to 150MHz\n");
		if(r = chase_da_SetClock(cardNum, 150000000)){
			fprintf(stderr, "Error unable to set clock to 150MHz: %d\n", r);
			return 1;		
		}
	}

	printf("Initialise Segment\n");
	// Create Segment #1
	for(i = 0; i < nPoints; i++){
		data0[i] = i * (2047 / nPoints);
	}
	segs[0].segmentNum = 0;
	segs[0].data = data0;
	segs[0].numPoints = nPoints;
	segs[0].numLoops = 0;
	segs[0].beginPadValue = 2047;
	segs[0].endPadValue = 2047;
	segs[0].trigEnable = 0;
	segs[0].nextSegNum = 0;
	
	printf("Create Segments\n");
	chase_da_CreateSegments(cardNum, 0xFF, 1, segs);

	printf("Set a marker at the start\n");
	chase_da_SetMarkers(cardNum, 64, 0xF, 0xF);

	printf("Start Running\n");
	chase_da_SetTriggerMode(cardNum, CHASE_DA_MODE_RUN, CHASE_DA_TRIG_POS);

//	chase_da_SetTriggerMode(cardNum, CHASE_DA_MODE_TRIGRUN, CHASE_DA_TRIG_POS);
//	chase_da_SetSoftTrigger(cardNum);

	pause();
	
	return 0;
}
