/*******************************************************************************
 *	Da11000_dll.cpp		DA11000 interface
 *				T.Barnaby,	BEAM Ltd,	2009-04-22
 *******************************************************************************
 * Copyright (c) 2009 Chase Scientific Company. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//
//   SOFTWARE HISTORY (DLL ONLY)
//
//   DATE        REV     DESCRIPTION
//   ------      -----   -------------
// ===========================================================================
//        *****  ORIGINAL HISTORY FOR AWG1000 DLL (up to 6/8/03) *****
// ===========================================================================
//   01/00/99    1.00    Initial Release
//   06/06/99    1.01    Added RAM_OFFSET read from EEPROM
//   06/07/99    1.02    Added MEMORY_OPTION read from EEPROM.
//                       Default is 1 MWord with/without EEPROM info.
//                       4 => 4 MWords. See AWG1200_Open for addr info.
//   08/10/99    1.03    Added Set_RAM_Offset(WORD RAM_Offset) for adjusting
//                       time delay for BOTH steady state as well as triggered
//                       waveforms.
//   03/08/00    1.04    Made changes to SetDDS(), AWG1200_SetClock, and
//                       AWG1200_SetTriggerMode for sub 60 MHz operation for
//                       ITT (John Coupe). Needs C145, C83 ==> 1K each for
//                       Rev. C. Basically sets DDS to 0 Hz for shutdowns.
//
//   08/01/00    1.05    Added "AWG1200_Create_Soft_Trigger(void)" so that
//                       hard trigger can be simulated by software. Also allows
//                       segment to complete if system is not reset.
//
//   08/04/00    1.06    Added "AWG1200_CreateSegments()" for complete control
//                       over single or mult-segment operation.
//
//   08/15/00    1.07    Fixed bug in "AWG1200_CreateSegments()". Needed to add
//                       another command to segment before triggered segment.
//                       Originally there was only one at end of pad. Now there
//                       is an additional cmd at end of data to catch RPT count
//                       which is defined by the previous segment definition.
//
//   09/14/00    1.08    Inverted Attenuator values and fixed 64.75=>63.74.
//
//   12/16/00    1.09    Added "AWG1200_UpdateSegmentCmds()" for faster command
//                       code updates.
//   09/01/01    1.10    Added Third Bin for Embedded Code
//
//   06-08-2003  1.2     Changed all awg1200 references to awg1000.
//
//                       Changed Default MEMORY_OPTION to 4 always
//
// ===========================================================================
//        *****  MODIFIED FOR DA4300 AFTER THIS POINT  *****
// ===========================================================================
//
//   02-07-2004  0.50    Start Engineering Modifications to convert AWG1000
//                       multi-card to DA4300 multi-card and awg1000 to DA4300.
//
//
// ===========================================================================
//        *****  MODIFIED FOR DA11000 AFTER THIS POINT  *****
// ===========================================================================
//
//   10-05-2004          Change filenames/function calls to da11000.
//
//   22-04-2009  2.1     Modified for use with Linux
//   22-04-2009  2.2     Removed RegisterWinDriver routine
//
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

// NEW ACTIONS REQUIRED
// 1) Add beg/end pad selection for single segment call.
// 2)
// 3)


#ifdef linux
#include <stdint.h>
#include "da11000_lib.h"
#else
#include <vcl.h>
//#pragma hdrstop

#include <windows.h>
#include <winioctl.h>
#include "windrvr.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#ifdef linux
#define EXPORT
#include <exception>
#include <iostream>
#else
#include <exception>
#include <iostream.h>

#include "p9030_lib.h"

#define EXPORT extern "C" __declspec(dllexport)
#endif

EXPORT DWORD da11000_CountCards(void);
EXPORT DWORD da11000_Open(DWORD CardNum);
EXPORT DWORD da11000_Close(DWORD CardNum);

EXPORT void da11000_SetClock(DWORD CardNum, DWORD Frequency);

EXPORT void da11000_SetTriggerMode(DWORD CardNum, BYTE Mode, BYTE ExtPol);
EXPORT void da11000_SetSoftTrigger(DWORD CardNum);
EXPORT void da11000_SetMarkers(DWORD CardNum, DWORD PointAddr, BYTE Nib1, BYTE Nib2);
EXPORT void da11000_SetOffset(DWORD CardNum, DWORD ChanNum, int Mode, int Offset);

EXPORT void da11000_CreateSingleSegment(DWORD CardNum, DWORD ChanNum, DWORD NumPoints, DWORD NumLoops, PVOID UserArrayPtr, DWORD TrigEn);
EXPORT void da11000_CreateSegments(DWORD CardNum, DWORD ChanNum, DWORD NumSegments, PVOID PtrToSegmentsList);
EXPORT void da11000_UpdateSegmentCmds(DWORD CardNum, DWORD ChanNum, DWORD NumSegments, PVOID PtrToSegmentsList);

EXPORT void da11000_Set_Atten(DWORD CardNum, DWORD ChanNum, DWORD Atten_Value);

//EXPORT void da11000_SetNoiseAtten(DOUBLE Atten_dB);
//EXPORT void da11000_SetMainAtten(DOUBLE Atten_dB);


#ifndef linux
//---------------------------------------------------------------------------
//  FOR DEBUG ONLY
//---------------------------------------------------------------------------

EXPORT void da11000_WritePoint(DWORD CardNum, BYTE BufferNum, DWORD PointAddr, WORD PointValue);
EXPORT WORD da11000_ReadPoint(DWORD CardNum, BYTE BufferNum, DWORD PointAddr);
EXPORT void da11000_WriteBlock(DWORD CardNum, BYTE BufferNum, DWORD StartAddr, DWORD Range, PVOID UserArrayPtr );
EXPORT void da11000_ReadBlock(DWORD CardNum, BYTE BufferNum, DWORD StartAddr, DWORD Range, PVOID UserArrayPtr );

EXPORT DWORD da11000_ReadPCIConfigReg(DWORD BrdNum, DWORD dwReg);
EXPORT void da11000_WritePCIConfigReg(DWORD BrdNum, DWORD dwReg, DWORD RegValue);
EXPORT DWORD da11000_ReadPCILocalReg(DWORD BrdNum, DWORD dwReg);
EXPORT void da11000_WritePCILocalReg(DWORD BrdNum, DWORD dwReg, DWORD RegValue);

EXPORT DWORD da11000_EEPROMReadDWord(DWORD BrdNum, DWORD dwReg);
EXPORT void da11000_EEPROMWriteDWord(DWORD BrdNum, DWORD dwReg, DWORD RegValue);

EXPORT void RunTest(DWORD CardNum);
EXPORT int Sine_64x(DWORD CardNum);
EXPORT int Sine_32x(DWORD CardNum);
EXPORT int Sine_16x(DWORD CardNum);
EXPORT int Sine_8x(DWORD CardNum);
EXPORT int Sine_4x(DWORD CardNum);
EXPORT int Zeros(DWORD CardNum);

EXPORT int MemTest(DWORD CardNum);
EXPORT int PlacePoint(DWORD CardNum);
EXPORT int PlacePoint2(DWORD CardNum);
EXPORT int PlacePoint3(DWORD CardNum);
EXPORT int PlacePoint4(DWORD CardNum);
#endif

// Start Local Calls
EXPORT DWORD RegisterWinDriver(void);
EXPORT void PCI_Bus_Enable(DWORD CardNum, BYTE PCI_SEL, BYTE Update);
EXPORT void Set_RAM_Offset(DWORD CardNum, WORD RAM_Offset);

EXPORT void SetDeviceID(DWORD NewDeviceID);

void EEPROM_DELAY(DWORD BrdNum);

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

HANDLE da11000_hWD = NULL;           // HANDLE TO WINDRIVER
P9030HANDLE da11000_hPlx = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx1 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx2 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx3 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx4 = NULL;     // HANDLE TO da11000 PLX9030 CARD

P9030HANDLE da11000_hPlx5 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx6 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx7 = NULL;     // HANDLE TO da11000 PLX9030 CARD
P9030HANDLE da11000_hPlx8 = NULL;     // HANDLE TO da11000 PLX9030 CARD

bool da11000_hPlx1_Initialized = false;
bool da11000_hPlx2_Initialized = false;
bool da11000_hPlx3_Initialized = false;
bool da11000_hPlx4_Initialized = false;
bool da11000_hPlx5_Initialized = false;
bool da11000_hPlx6_Initialized = false;
bool da11000_hPlx7_Initialized = false;
bool da11000_hPlx8_Initialized = false;

P9030HANDLE ATTEN_hPlx   = NULL;     // HANDLE TO ATTEN PLX9030 CARD
P9030HANDLE Test_hPlx   = NULL;     // HANDLE TO ATTEN PLX9030 CARD

DWORD NumCards = 0;


enum {
    R0 = 0,     //  da11000 R/W CONTROL REGISTERS OFFSETS FROM
    W0 = 0,     //  ADDRESS SPACE1
    W1 = 2,     //

    W2 = 4,     //  LTCH_RPT_CNT
    W3 = 6,     //  LTCH_LW_LIMIT
    W4 = 8      //  LTCH_UW_LIMIT
};

WORD WGR[3]= {0x499F,0x0041};   // Create and initialize ghost registers.
DWORD LastCard = 0;


int ClockRate;
BYTE PCI_SEL_MODE;  // Keeps track of current internal PCI Bus direction

// OUTPUT OFFSET VARIABLES
WORD Offset_Max1, Offset_Zero1, Offset_Min1;
WORD Offset_Max2, Offset_Zero2, Offset_Min2;
WORD Offset_Max3, Offset_Zero3, Offset_Min3;
WORD Offset_Max4, Offset_Zero4, Offset_Min4;
WORD Offset_Max5, Offset_Zero5, Offset_Min5;
WORD Offset_Max6, Offset_Zero6, Offset_Min6;
WORD Offset_Max7, Offset_Zero7, Offset_Min7;
WORD Offset_Max8, Offset_Zero8, Offset_Min8;
WORD Last_DAC_A_Value, Last_DAC_B_Value, Last_DAC_C_Value, Last_DAC_D_Value;

// RAM OFFSET INCREMENT
WORD RAM_OFFSET[100];

// MISC
WORD MEMORY_OPTION = 4;  // 1 MEG MEMORY DEFAULT
WORD Arb_Location = 0;

DWORD DeviceID = 0x9031;  // Default for DA11000-12-4M-PCI
DWORD VendorID = 0x10B5;

#ifndef linux
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void*)
{
    return 1;
}

//---------------------------------------------------------------------------

void SetDeviceID(DWORD NewDeviceID)  {
    DeviceID = 0x0000FFFF & NewDeviceID;
    VendorID = NewDeviceID >> 16;
}

//---------------------------------------------------------------------------

DWORD RegisterWinDriver()
{
	return -1;
}
#endif

//---------------------------------------------------------------------------

DWORD da11000_CountCards(void)
{
#ifdef linux
	return P9030_CountCards();
#else
    DWORD Test = RegisterWinDriver();  // Registers WinDriver using latest 6.03 License
    if (Test == 0) return(P9030_CountCards(VendorID, DeviceID));
    else return(13);   // Let da11000_Open know that Kernel Driver is not working
#endif
}

//---------------------------------------------------------------------------

WORD WGR_Brd1[3]= {0x499F,0x0041};
DWORD LastCard_Brd1 = 0;
int ClockRate_Brd1;
BYTE PCI_SEL_MODE_Brd1;
WORD Offset_Max1_Brd1, Offset_Zero1_Brd1, Offset_Min1_Brd1;
WORD Offset_Max2_Brd1, Offset_Zero2_Brd1, Offset_Min2_Brd1;
WORD Last_DAC_A_Value_Brd1, Last_DAC_B_Value_Brd1, Last_DAC_C_Value_Brd1, Last_DAC_D_Value_Brd1;
WORD RAM_OFFSET_Brd1[100];
WORD MEMORY_OPTION_Brd1 = 4;
WORD Arb_Location_Brd1 = 0;

WORD WGR_Brd2[3]= {0x499F,0x0041};
DWORD LastCard_Brd2 = 0;
int ClockRate_Brd2;
BYTE PCI_SEL_MODE_Brd2;
WORD Offset_Max1_Brd2, Offset_Zero1_Brd2, Offset_Min1_Brd2;
WORD Offset_Max2_Brd2, Offset_Zero2_Brd2, Offset_Min2_Brd2;
WORD Last_DAC_A_Value_Brd2, Last_DAC_B_Value_Brd2, Last_DAC_C_Value_Brd2, Last_DAC_D_Value_Brd2;
WORD RAM_OFFSET_Brd2[100];
WORD MEMORY_OPTION_Brd2 = 4;
WORD Arb_Location_Brd2 = 0;

WORD WGR_Brd3[3]= {0x499F,0x0041};
DWORD LastCard_Brd3 = 0;
int ClockRate_Brd3;
BYTE PCI_SEL_MODE_Brd3;
WORD Offset_Max1_Brd3, Offset_Zero1_Brd3, Offset_Min1_Brd3;
WORD Offset_Max2_Brd3, Offset_Zero2_Brd3, Offset_Min2_Brd3;
WORD Last_DAC_A_Value_Brd3, Last_DAC_B_Value_Brd3, Last_DAC_C_Value_Brd3, Last_DAC_D_Value_Brd3;
WORD RAM_OFFSET_Brd3[100];
WORD MEMORY_OPTION_Brd3 = 4;
WORD Arb_Location_Brd3 = 0;

WORD WGR_Brd4[3]= {0x499F,0x0041};
DWORD LastCard_Brd4 = 0;
int ClockRate_Brd4;
BYTE PCI_SEL_MODE_Brd4;
WORD Offset_Max1_Brd4, Offset_Zero1_Brd4, Offset_Min1_Brd4;
WORD Offset_Max2_Brd4, Offset_Zero2_Brd4, Offset_Min2_Brd4;
WORD Last_DAC_A_Value_Brd4, Last_DAC_B_Value_Brd4, Last_DAC_C_Value_Brd4, Last_DAC_D_Value_Brd4;
WORD RAM_OFFSET_Brd4[100];
WORD MEMORY_OPTION_Brd4 = 4;
WORD Arb_Location_Brd4 = 0;

// Additional Boards 10/25/05 for GAGE (required 6 total, added 8 total)

WORD WGR_Brd5[3]= {0x499F,0x0041};
DWORD LastCard_Brd5 = 0;
int ClockRate_Brd5;
BYTE PCI_SEL_MODE_Brd5;
WORD Offset_Max1_Brd5, Offset_Zero1_Brd5, Offset_Min1_Brd5;
WORD Offset_Max2_Brd5, Offset_Zero2_Brd5, Offset_Min2_Brd5;
WORD Last_DAC_A_Value_Brd5, Last_DAC_B_Value_Brd5, Last_DAC_C_Value_Brd5, Last_DAC_D_Value_Brd5;
WORD RAM_OFFSET_Brd5[100];
WORD MEMORY_OPTION_Brd5 = 4;
WORD Arb_Location_Brd5 = 0;

WORD WGR_Brd6[3]= {0x499F,0x0041};
DWORD LastCard_Brd6 = 0;
int ClockRate_Brd6;
BYTE PCI_SEL_MODE_Brd6;
WORD Offset_Max1_Brd6, Offset_Zero1_Brd6, Offset_Min1_Brd6;
WORD Offset_Max2_Brd6, Offset_Zero2_Brd6, Offset_Min2_Brd6;
WORD Last_DAC_A_Value_Brd6, Last_DAC_B_Value_Brd6, Last_DAC_C_Value_Brd6, Last_DAC_D_Value_Brd6;
WORD RAM_OFFSET_Brd6[100];
WORD MEMORY_OPTION_Brd6 = 4;
WORD Arb_Location_Brd6 = 0;

WORD WGR_Brd7[3]= {0x499F,0x0041};
DWORD LastCard_Brd7 = 0;
int ClockRate_Brd7;
BYTE PCI_SEL_MODE_Brd7;
WORD Offset_Max1_Brd7, Offset_Zero1_Brd7, Offset_Min1_Brd7;
WORD Offset_Max2_Brd7, Offset_Zero2_Brd7, Offset_Min2_Brd7;
WORD Last_DAC_A_Value_Brd7, Last_DAC_B_Value_Brd7, Last_DAC_C_Value_Brd7, Last_DAC_D_Value_Brd7;
WORD RAM_OFFSET_Brd7[100];
WORD MEMORY_OPTION_Brd7 = 4;
WORD Arb_Location_Brd7 = 0;

WORD WGR_Brd8[3]= {0x499F,0x0041};
DWORD LastCard_Brd8 = 0;
int ClockRate_Brd8;
BYTE PCI_SEL_MODE_Brd8;
WORD Offset_Max1_Brd8, Offset_Zero1_Brd8, Offset_Min1_Brd8;
WORD Offset_Max2_Brd8, Offset_Zero2_Brd8, Offset_Min2_Brd8;
WORD Last_DAC_A_Value_Brd8, Last_DAC_B_Value_Brd8, Last_DAC_C_Value_Brd8, Last_DAC_D_Value_Brd8;
WORD RAM_OFFSET_Brd8[100];
WORD MEMORY_OPTION_Brd8 = 4;
WORD Arb_Location_Brd8 = 0;


void BackupRegs(DWORD CardNum)
{
    int x;

    if (CardNum == 1)
    {
        for (x = 0; x < 3; x++) WGR_Brd1[x] = WGR[x];
//        LastCard_Brd1 = LastCard;

        ClockRate_Brd1 = ClockRate;
        PCI_SEL_MODE_Brd1 = PCI_SEL_MODE;

        Offset_Max1_Brd1 = Offset_Max1; Offset_Zero1_Brd1 = Offset_Zero1; Offset_Min1_Brd1 = Offset_Min1;
        Offset_Max2_Brd1 = Offset_Max2; Offset_Zero2_Brd1 = Offset_Zero2; Offset_Min2_Brd1 = Offset_Min2;

        Last_DAC_A_Value_Brd1 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd1 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd1 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd1 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd1[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd1 = MEMORY_OPTION;
        Arb_Location_Brd1 = Arb_Location;
    }

    if (CardNum == 2)
    {
        for (x = 0; x < 3; x++) WGR_Brd2[x] = WGR[x];
//        LastCard_Brd2 = LastCard;

        ClockRate_Brd2 = ClockRate;
        PCI_SEL_MODE_Brd2 = PCI_SEL_MODE;

        Offset_Max1_Brd2 = Offset_Max1; Offset_Zero1_Brd2 = Offset_Zero1; Offset_Min1_Brd2 = Offset_Min1;
        Offset_Max2_Brd2 = Offset_Max2; Offset_Zero2_Brd2 = Offset_Zero2; Offset_Min2_Brd2 = Offset_Min2;

        Last_DAC_A_Value_Brd2 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd2 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd2 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd2 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd2[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd2 = MEMORY_OPTION;
        Arb_Location_Brd2 = Arb_Location;
    }

    if (CardNum == 3)
    {
        for (x = 0; x < 3; x++) WGR_Brd3[x] = WGR[x];
//        LastCard_Brd3 = LastCard;

        ClockRate_Brd3 = ClockRate;
        PCI_SEL_MODE_Brd3 = PCI_SEL_MODE;

        Offset_Max1_Brd3 = Offset_Max1; Offset_Zero1_Brd3 = Offset_Zero1; Offset_Min1_Brd3 = Offset_Min1;
        Offset_Max2_Brd3 = Offset_Max2; Offset_Zero2_Brd3 = Offset_Zero2; Offset_Min2_Brd3 = Offset_Min2;

        Last_DAC_A_Value_Brd3 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd3 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd3 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd3 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd3[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd3 = MEMORY_OPTION;
        Arb_Location_Brd3 = Arb_Location;
    }

    if (CardNum == 4)
    {
        for (x = 0; x < 3; x++) WGR_Brd4[x] = WGR[x];
//        LastCard_Brd4 = LastCard;

        ClockRate_Brd4 = ClockRate;
        PCI_SEL_MODE_Brd4 = PCI_SEL_MODE;

        Offset_Max1_Brd4 = Offset_Max1; Offset_Zero1_Brd4 = Offset_Zero1; Offset_Min1_Brd4 = Offset_Min1;
        Offset_Max2_Brd4 = Offset_Max2; Offset_Zero2_Brd4 = Offset_Zero2; Offset_Min2_Brd4 = Offset_Min2;

        Last_DAC_A_Value_Brd4 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd4 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd4 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd4 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd4[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd4 = MEMORY_OPTION;
        Arb_Location_Brd4 = Arb_Location;
    }

// NEW ==> Added 4 more

    if (CardNum == 5)
    {
        for (x = 0; x < 3; x++) WGR_Brd5[x] = WGR[x];
//        LastCard_Brd5 = LastCard;

        ClockRate_Brd5 = ClockRate;
        PCI_SEL_MODE_Brd5 = PCI_SEL_MODE;

        Offset_Max1_Brd5 = Offset_Max1; Offset_Zero1_Brd5 = Offset_Zero1; Offset_Min1_Brd5 = Offset_Min1;
        Offset_Max2_Brd5 = Offset_Max2; Offset_Zero2_Brd5 = Offset_Zero2; Offset_Min2_Brd5 = Offset_Min2;

        Last_DAC_A_Value_Brd5 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd5 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd5 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd5 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd5[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd5 = MEMORY_OPTION;
        Arb_Location_Brd5 = Arb_Location;
    }

    if (CardNum == 6)
    {
        for (x = 0; x < 3; x++) WGR_Brd6[x] = WGR[x];
//        LastCard_Brd6 = LastCard;

        ClockRate_Brd6 = ClockRate;
        PCI_SEL_MODE_Brd6 = PCI_SEL_MODE;

        Offset_Max1_Brd6 = Offset_Max1; Offset_Zero1_Brd6 = Offset_Zero1; Offset_Min1_Brd6 = Offset_Min1;
        Offset_Max2_Brd6 = Offset_Max2; Offset_Zero2_Brd6 = Offset_Zero2; Offset_Min2_Brd6 = Offset_Min2;

        Last_DAC_A_Value_Brd6 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd6 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd6 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd6 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd6[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd6 = MEMORY_OPTION;
        Arb_Location_Brd6 = Arb_Location;
    }

    if (CardNum == 7)
    {
        for (x = 0; x < 3; x++) WGR_Brd7[x] = WGR[x];
//        LastCard_Brd7 = LastCard;

        ClockRate_Brd7 = ClockRate;
        PCI_SEL_MODE_Brd7 = PCI_SEL_MODE;

        Offset_Max1_Brd7 = Offset_Max1; Offset_Zero1_Brd7 = Offset_Zero1; Offset_Min1_Brd7 = Offset_Min1;
        Offset_Max2_Brd7 = Offset_Max2; Offset_Zero2_Brd7 = Offset_Zero2; Offset_Min2_Brd7 = Offset_Min2;

        Last_DAC_A_Value_Brd7 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd7 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd7 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd7 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd7[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd7 = MEMORY_OPTION;
        Arb_Location_Brd7 = Arb_Location;
    }

    if (CardNum == 8)
    {
        for (x = 0; x < 3; x++) WGR_Brd8[x] = WGR[x];
//        LastCard_Brd8 = LastCard;

        ClockRate_Brd8 = ClockRate;
        PCI_SEL_MODE_Brd8 = PCI_SEL_MODE;

        Offset_Max1_Brd8 = Offset_Max1; Offset_Zero1_Brd8 = Offset_Zero1; Offset_Min1_Brd8 = Offset_Min1;
        Offset_Max2_Brd8 = Offset_Max2; Offset_Zero2_Brd8 = Offset_Zero2; Offset_Min2_Brd8 = Offset_Min2;

        Last_DAC_A_Value_Brd8 = Last_DAC_A_Value;
        Last_DAC_B_Value_Brd8 = Last_DAC_B_Value;
        Last_DAC_C_Value_Brd8 = Last_DAC_C_Value;
        Last_DAC_D_Value_Brd8 = Last_DAC_D_Value;

        for (x = 0; x < 100; x++) RAM_OFFSET_Brd8[x] = RAM_OFFSET[x];
        MEMORY_OPTION_Brd8 = MEMORY_OPTION;
        Arb_Location_Brd8 = Arb_Location;
    }



}

void RestoreRegs(DWORD CardNum)
{
    int x;

    if (CardNum == 1)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd1[x];
        LastCard = LastCard_Brd1;

        ClockRate = ClockRate_Brd1;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd1;

        Offset_Max1 = Offset_Max1_Brd1; Offset_Zero1 = Offset_Zero1_Brd1; Offset_Min1 = Offset_Min1_Brd1;
        Offset_Max2 = Offset_Max2_Brd1; Offset_Zero2 = Offset_Zero2_Brd1; Offset_Min2 = Offset_Min2_Brd1;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd1;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd1;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd1;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd1;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd1[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd1;
        Arb_Location = Arb_Location_Brd1;
    }

    if (CardNum == 2)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd2[x];
        LastCard = LastCard_Brd2;

        ClockRate = ClockRate_Brd2;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd2;

        Offset_Max1 = Offset_Max1_Brd2; Offset_Zero1 = Offset_Zero1_Brd2; Offset_Min1 = Offset_Min1_Brd2;
        Offset_Max2 = Offset_Max2_Brd2; Offset_Zero2 = Offset_Zero2_Brd2; Offset_Min2 = Offset_Min2_Brd2;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd2;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd2;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd2;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd2;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd2[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd2;
        Arb_Location = Arb_Location_Brd2;
    }

    if (CardNum == 3)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd3[x];
        LastCard = LastCard_Brd3;

        ClockRate = ClockRate_Brd3;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd3;

        Offset_Max1 = Offset_Max1_Brd3; Offset_Zero1 = Offset_Zero1_Brd3; Offset_Min1 = Offset_Min1_Brd3;
        Offset_Max2 = Offset_Max2_Brd3; Offset_Zero2 = Offset_Zero2_Brd3; Offset_Min2 = Offset_Min2_Brd3;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd3;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd3;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd3;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd3;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd3[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd3;
        Arb_Location = Arb_Location_Brd3;
    }

    if (CardNum == 4)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd4[x];
        LastCard = LastCard_Brd4;

        ClockRate = ClockRate_Brd4;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd4;

        Offset_Max1 = Offset_Max1_Brd4; Offset_Zero1 = Offset_Zero1_Brd4; Offset_Min1 = Offset_Min1_Brd4;
        Offset_Max2 = Offset_Max2_Brd4; Offset_Zero2 = Offset_Zero2_Brd4; Offset_Min2 = Offset_Min2_Brd4;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd4;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd4;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd4;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd4;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd4[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd4;
        Arb_Location = Arb_Location_Brd4;
    }

// NEW ==> Added 4 more

    if (CardNum == 5)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd5[x];
        LastCard = LastCard_Brd5;

        ClockRate = ClockRate_Brd5;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd5;

        Offset_Max1 = Offset_Max1_Brd5; Offset_Zero1 = Offset_Zero1_Brd5; Offset_Min1 = Offset_Min1_Brd5;
        Offset_Max2 = Offset_Max2_Brd5; Offset_Zero2 = Offset_Zero2_Brd5; Offset_Min2 = Offset_Min2_Brd5;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd5;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd5;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd5;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd5;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd5[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd5;
        Arb_Location = Arb_Location_Brd5;
    }

    if (CardNum == 6)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd6[x];
        LastCard = LastCard_Brd6;

        ClockRate = ClockRate_Brd6;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd6;

        Offset_Max1 = Offset_Max1_Brd6; Offset_Zero1 = Offset_Zero1_Brd6; Offset_Min1 = Offset_Min1_Brd6;
        Offset_Max2 = Offset_Max2_Brd6; Offset_Zero2 = Offset_Zero2_Brd6; Offset_Min2 = Offset_Min2_Brd6;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd6;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd6;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd6;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd6;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd6[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd6;
        Arb_Location = Arb_Location_Brd6;
    }

    if (CardNum == 7)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd7[x];
        LastCard = LastCard_Brd7;

        ClockRate = ClockRate_Brd7;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd7;

        Offset_Max1 = Offset_Max1_Brd7; Offset_Zero1 = Offset_Zero1_Brd7; Offset_Min1 = Offset_Min1_Brd7;
        Offset_Max2 = Offset_Max2_Brd7; Offset_Zero2 = Offset_Zero2_Brd7; Offset_Min2 = Offset_Min2_Brd7;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd7;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd7;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd7;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd7;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd7[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd7;
        Arb_Location = Arb_Location_Brd7;
    }

    if (CardNum == 8)
    {
        for (x = 0; x < 3; x++) WGR[x] = WGR_Brd8[x];
        LastCard = LastCard_Brd8;

        ClockRate = ClockRate_Brd8;
        PCI_SEL_MODE = PCI_SEL_MODE_Brd8;

        Offset_Max1 = Offset_Max1_Brd8; Offset_Zero1 = Offset_Zero1_Brd8; Offset_Min1 = Offset_Min1_Brd8;
        Offset_Max2 = Offset_Max2_Brd8; Offset_Zero2 = Offset_Zero2_Brd8; Offset_Min2 = Offset_Min2_Brd8;

        Last_DAC_A_Value = Last_DAC_A_Value_Brd8;
        Last_DAC_B_Value = Last_DAC_B_Value_Brd8;
        Last_DAC_C_Value = Last_DAC_C_Value_Brd8;
        Last_DAC_D_Value = Last_DAC_D_Value_Brd8;

        for (x = 0; x < 100; x++) RAM_OFFSET[x] = RAM_OFFSET_Brd8[x];
        MEMORY_OPTION = MEMORY_OPTION_Brd8;
        Arb_Location = Arb_Location_Brd8;
    }




}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

bool BrdNumOK(DWORD CardNum)
{
    bool ReturnValue = false;

    if ((CardNum == 1) && (da11000_hPlx1 != NULL)) ReturnValue = true;
    if ((CardNum == 2) && (da11000_hPlx2 != NULL)) ReturnValue = true;
    if ((CardNum == 3) && (da11000_hPlx3 != NULL)) ReturnValue = true;
    if ((CardNum == 4) && (da11000_hPlx4 != NULL)) ReturnValue = true;
    if ((CardNum == 5) && (da11000_hPlx5 != NULL)) ReturnValue = true;
    if ((CardNum == 6) && (da11000_hPlx6 != NULL)) ReturnValue = true;
    if ((CardNum == 7) && (da11000_hPlx7 != NULL)) ReturnValue = true;
    if ((CardNum == 8) && (da11000_hPlx8 != NULL)) ReturnValue = true;

    if ((ReturnValue == true) && (LastCard != CardNum) && (LastCard != 0))
    {
        BackupRegs(LastCard);
        switch (CardNum) {
           case 1: if (da11000_hPlx1_Initialized) RestoreRegs(1); break;   // only restore if already initialized
           case 2: if (da11000_hPlx2_Initialized) RestoreRegs(2); break;   // only restore if already initialized
           case 3: if (da11000_hPlx3_Initialized) RestoreRegs(3); break;   // only restore if already initialized
           case 4: if (da11000_hPlx4_Initialized) RestoreRegs(4); break;   // only restore if already initialized
           case 5: if (da11000_hPlx4_Initialized) RestoreRegs(5); break;   // only restore if already initialized
           case 6: if (da11000_hPlx4_Initialized) RestoreRegs(6); break;   // only restore if already initialized
           case 7: if (da11000_hPlx4_Initialized) RestoreRegs(7); break;   // only restore if already initialized
           case 8: if (da11000_hPlx4_Initialized) RestoreRegs(8); break;   // only restore if already initialized
        }
        LastCard = CardNum;

//        if ((LastCard == 1) && (CardNum == 2))
//        {
//            BackupRegs(1);
//            if (da11000_hPlx2_Initialized) RestoreRegs(2); // only restore if already initialized
//            LastCard = 2;
//        }
//        if ((LastCard == 2) && (CardNum == 1))
//        {
//            BackupRegs(2);
//            if (da11000_hPlx1_Initialized) RestoreRegs(1); // only restore if already initialized
//            LastCard = 1;
//        }
    }

    if ((ReturnValue == true) && (LastCard != CardNum) && (LastCard == 0))
    {
        if (CardNum == 1) LastCard = 1;
        if (CardNum == 2) LastCard = 2;
        if (CardNum == 3) LastCard = 3;
        if (CardNum == 4) LastCard = 4;
        if (CardNum == 5) LastCard = 5;
        if (CardNum == 6) LastCard = 6;
        if (CardNum == 7) LastCard = 7;
        if (CardNum == 8) LastCard = 8;
    }

    return(ReturnValue);
}

P9030HANDLE GetBrdHandle(DWORD CardNum)
{
    if ((CardNum == 1) && (da11000_hPlx1 != NULL)) return(da11000_hPlx1);
    if ((CardNum == 2) && (da11000_hPlx2 != NULL)) return(da11000_hPlx2);
    if ((CardNum == 3) && (da11000_hPlx3 != NULL)) return(da11000_hPlx3);
    if ((CardNum == 4) && (da11000_hPlx4 != NULL)) return(da11000_hPlx4);
    if ((CardNum == 5) && (da11000_hPlx5 != NULL)) return(da11000_hPlx5);
    if ((CardNum == 6) && (da11000_hPlx6 != NULL)) return(da11000_hPlx6);
    if ((CardNum == 7) && (da11000_hPlx7 != NULL)) return(da11000_hPlx7);
    if ((CardNum == 8) && (da11000_hPlx8 != NULL)) return(da11000_hPlx8);

    return(NULL);
}

//---------------------------------------------------------------------------

void WritePort(WORD PortNum, WORD PortValue)    // Actually Memory Write in this case
{
    if (da11000_hPlx == NULL) return;


    switch (PortNum) {
        case 0:   // WRITE PORT 0
            WGR[0] = PortValue;
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W0, WGR[0]);
            break;
        case 1:   // WRITE PORT 1
            WGR[1] = PortValue;
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W1, WGR[1]);
            break;
        case 2:   // LTCH_RPT_CNT
            WGR[2] = PortValue;
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W2, WGR[2]);
            break;
        case 3:   // LTCH_LW_LIMIT
            WGR[3] = PortValue;
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W3, WGR[3]);
            break;
        case 4:   // LTCH_UW_LIMIT
            WGR[4] = PortValue;
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W4, WGR[4]);
            break;
    }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

void Load_PLL(DWORD BrdNum, DWORD RegValue)  // LMX2316
{
//  LSB                                                       SHIFTED IN FIRST ==>  MSB
//  C1  C2  F1  F2  F3  F4  F5  F6  F7  F8  F9  F10 F11 F12 F13 F14 F15 F16 F17 F18 F19

    WritePort(0, WGR[0] & 0xFFFC);  // Set SER_DATA & SER_CLOCK to zero
EEPROM_DELAY(BrdNum);
    WritePort(1, WGR[1] & 0xFFEF);  // Set PLL_LD_EN to Zero - Allows clocking data into PLL
EEPROM_DELAY(BrdNum);

    for (DWORD i=0; i < 21; i++) {
        if ((RegValue & 0x100000) > 0) WritePort(0, WGR[0] | 0x0001);   // SET DATA VALUE
        else WritePort(0, WGR[0] & 0xFFFE);

EEPROM_DELAY(BrdNum);
        WritePort(0, WGR[0] | 0x0002); WritePort(0, WGR[0] & 0xFFFD);  // CLOCK DATA
        RegValue = RegValue << 1;                                      // Shift left to Grab Next Bit
    }

EEPROM_DELAY(BrdNum);
    WritePort(1, WGR[1] | 0x0010);  // LOADS REGISTER VALUES INTO DEVICE
EEPROM_DELAY(BrdNum);

}

//---------------------------------------------------------------------------

DWORD da11000_Open(DWORD CardNum)
{
    DWORD RAM_Offset; DWORD CLK_Offset;

    NumCards = da11000_CountCards();
    if (NumCards == 13) return(13);       // Abort further tests if Kernel Driver not working
                                          // Also report with error code 13.

    DWORD ReturnCode = 1;  // Default as "No P9030 cards detected" if no other code determined

    if (NumCards == 0) return(ReturnCode);         // Kernel Load ok, but no cards detected

    if (NumCards > 0){
        if (CardNum > NumCards) ReturnCode = 6;  //  Card number exceeds number of cards

         switch (CardNum) {
            case 1: if (da11000_hPlx1 != NULL) ReturnCode = 3; break;    // Board already open
            case 2: if (da11000_hPlx2 != NULL) ReturnCode = 3; break;
            case 3: if (da11000_hPlx3 != NULL) ReturnCode = 3; break;
            case 4: if (da11000_hPlx4 != NULL) ReturnCode = 3; break;
            case 5: if (da11000_hPlx5 != NULL) ReturnCode = 3; break;
            case 6: if (da11000_hPlx6 != NULL) ReturnCode = 3; break;
            case 7: if (da11000_hPlx7 != NULL) ReturnCode = 3; break;
            case 8: if (da11000_hPlx8 != NULL) ReturnCode = 3; break;
        }

        if (ReturnCode != 3)
        {
            switch (CardNum) {
               case 1:  if (P9030_Open (&da11000_hPlx1, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 2:  if (P9030_Open (&da11000_hPlx2, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 3:  if (P9030_Open (&da11000_hPlx3, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 4:  if (P9030_Open (&da11000_hPlx4, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 5:  if (P9030_Open (&da11000_hPlx5, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 6:  if (P9030_Open (&da11000_hPlx6, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 7:  if (P9030_Open (&da11000_hPlx7, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
               case 8:  if (P9030_Open (&da11000_hPlx8, VendorID, DeviceID, CardNum-1)) {ReturnCode = 0; break;}
                            else ReturnCode = 2; break;
            }
        }
    }

    if (ReturnCode > 0) return(ReturnCode);  // Stop here if error occurred.

    switch (CardNum) {
       case 1:  da11000_hPlx = da11000_hPlx1; break;    // Assign generic handle
       case 2:  da11000_hPlx = da11000_hPlx2; break;
       case 3:  da11000_hPlx = da11000_hPlx3; break;
       case 4:  da11000_hPlx = da11000_hPlx4; break;
       case 5:  da11000_hPlx = da11000_hPlx5; break;
       case 6:  da11000_hPlx = da11000_hPlx6; break;
       case 7:  da11000_hPlx = da11000_hPlx7; break;
       case 8:  da11000_hPlx = da11000_hPlx8; break;
    }

    if (!BrdNumOK(CardNum)) return(11);     // Dummy call Swaps Ghost Regs if necessary


//----- INITIALIZE CARD IF FIRST TIME --------------------------------------

// LMX2316 PROGRAMMING
// LOAD FUNCTION LATCH (0x10), THEN R COUNTER (0x00), THEN N COUNTER (0x01)

    DWORD Function_LTCH = 0x000092;   // 0x000092
    DWORD R_Counter     = 0x100028;   // 0x100050 => 20MHz, 0x100028 => 10MHz
    DWORD N_Counter     = 0x000FA1;   // 0x100FA1


    if ((CardNum == 1) && (da11000_hPlx1_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 1) && (!da11000_hPlx1_Initialized)) da11000_hPlx1_Initialized = true;

    if ((CardNum == 2) && (da11000_hPlx2_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 2) && (!da11000_hPlx2_Initialized)) da11000_hPlx2_Initialized = true;

    if ((CardNum == 3) && (da11000_hPlx3_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 3) && (!da11000_hPlx3_Initialized)) da11000_hPlx3_Initialized = true;

    if ((CardNum == 4) && (da11000_hPlx4_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 4) && (!da11000_hPlx4_Initialized)) da11000_hPlx4_Initialized = true;


    if ((CardNum == 5) && (da11000_hPlx5_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 5) && (!da11000_hPlx5_Initialized)) da11000_hPlx5_Initialized = true;

    if ((CardNum == 6) && (da11000_hPlx6_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 6) && (!da11000_hPlx6_Initialized)) da11000_hPlx6_Initialized = true;

    if ((CardNum == 7) && (da11000_hPlx7_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 7) && (!da11000_hPlx7_Initialized)) da11000_hPlx7_Initialized = true;

    if ((CardNum == 8) && (da11000_hPlx8_Initialized)) goto bypass;  // Initialize variables only once.
    if ((CardNum == 8) && (!da11000_hPlx8_Initialized)) da11000_hPlx8_Initialized = true;


    WGR[0] = 0x4900;  // Initialize ghost registors
    WGR[1] = 0x0201;  // Sync Reset, Set RAM_Offset=2   (x | 0x0200)

    // void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W0, WGR[0]);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W1, WGR[1]);

    Load_PLL(CardNum, Function_LTCH);
    Load_PLL(CardNum,     R_Counter);
    Load_PLL(CardNum,     N_Counter);


// ==> SET CLOCK RATE HERE

    PCI_Bus_Enable(CardNum,0,1);   // PUT INTO ARB MODE BY DEFAULT - BETTER STARTUP TIMES; Update Global
    da11000_SetTriggerMode(CardNum,0,0);  // Initialize?


// ******************************************* //
//       INITIALIZE RAM_OFFSET / CLK_OFFSET    //
// ******************************************* //

//    RAM_Offset = (da11000_EEPROMReadDWord(CardNum,0x88) & 0xF);
//    if ( RAM_Offset > 0 )
        Set_RAM_Offset(CardNum, 0x3); //RAM_Offset);

//    CLK_Offset = (da11000_EEPROMReadDWord(CardNum,0x88) & 0x30);
//    if ( CLK_Offset > 0 ) {
        WritePort(0,(WGR[0] & 0xFFCF));       // Clear CLK_Offset bits
        WritePort(0,(WGR[0] | 0x10));  //CLK_Offset));   // Apply CLK_Offset  //==> default is 0x10
//    }


// ******************************** //
//       READ MEMORY OPTION         //
// ******************************** //

//    if (Read_E2Prom(CardNum,6) != 0xFFFF) MEMORY_OPTION = Read_E2Prom(CardNum,6); else MEMORY_OPTION = 1;
MEMORY_OPTION = 4;

bypass:    // Bypasses Initialization if not first time.

MEMORY_OPTION = 4;

    return(ReturnCode);
}

//---------------------------------------------------------------------------

DWORD da11000_Close(DWORD CardNum)
{
    if (NumCards == 13) return(13);       // Abort further tests if Kernel Driver not working
                                          // Also report with error code 13.

    if (!BrdNumOK(CardNum)) return(5);    // Board Already Closed
    da11000_hPlx = GetBrdHandle(CardNum);

    BackupRegs(CardNum);  // Keep clean backup of any card closed

    P9030_Close(da11000_hPlx);

    if (CardNum == 1) da11000_hPlx1 = NULL;
    if (CardNum == 2) da11000_hPlx2 = NULL;
    if (CardNum == 3) da11000_hPlx3 = NULL;
    if (CardNum == 4) da11000_hPlx4 = NULL;
    if (CardNum == 5) da11000_hPlx5 = NULL;
    if (CardNum == 6) da11000_hPlx6 = NULL;
    if (CardNum == 7) da11000_hPlx7 = NULL;
    if (CardNum == 8) da11000_hPlx8 = NULL;

    return(0);
}

//---------------------------------------------------------------------------

void da11000_SetClock(DWORD CardNum, DWORD Frequency)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

//    if (Frequency > 300000000) {
//        WritePort( 0,(WGR[0] & 0xF03F) | 0x0800 );  // Set Clock To External For Debug + Shut down TTL Clk
//    }
//    if (Frequency == 300000000) {
//        WritePort( 0,(WGR[0] & 0xF03F) | 0x0400 );  // Set Clock To Internal 300 MHz + DIV/1
//    }
//    if (Frequency == 150000000) {
//        WritePort( 0,(WGR[0] & 0xF03F) | 0x0040 );  // Set Clock To Internal 150 MHz + DIV/2
//    }
//    if (Frequency ==  75000000) {
//        WritePort( 0,(WGR[0] & 0xF03F) | 0x0080 );  // Set Clock To Internal 75 MHz + DIV/4
//    }
//

//    if (Frequency >= 1500e6) {
//        WritePort( 0,(WGR[0] & 0xF03F) | 0x800 );  // Set Clock Mux to External Clock + DIV/1
//        return;
//    }

    ClockRate = Frequency;

}

//---------------------------------------------------------------------------

void da11000_SetTriggerMode(DWORD CardNum, BYTE Mode, BYTE ExtPol)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if (Mode == 0)
    {
        WritePort(1,WGR[1] | 0x1);       // Shuts down MAD Counter
        WritePort( 0,WGR[0] & 0x7FFF);  // Shuts down TRIG_EN if on

        WritePort( 1,WGR[1] | 0x0080);  // SYNC_DAC_TTL
        WritePort( 1,WGR[1] & 0xEFFF);  // Disables VCO
// if (ClockRate < 60000000) SetDDS(0);  // Disables DDS if selected
     }

    if ( (Mode == 1) | (Mode == 2) )
    {
    // SYNCHRONIZE UPCONVERTERS
        WritePort(1,WGR[1] | 0x1);       // Shuts down MAD Counter

        WritePort( 1,WGR[1] & 0xEFFF);  // Disables VCO
// if (ClockRate < 60000000) SetDDS(0);  // Disables DDS if selected

        WritePort( 0,WGR[0] | 0x2000);  // SYNC_RAM_TTL
        WritePort( 0,WGR[0] & 0xDFFF);  // Toggles High to Reset

        WritePort( 1,WGR[1] | 0x0080);  // SYNC_DAC_TTL
        WritePort( 1,WGR[1] & 0xFF7F);  // Toggles HIGH to Reset


        if (Mode == 2) {
            WritePort( 0,WGR[0] | 0x8000);  // Activates TRIG_EN
            if (ExtPol == 1) {
              WritePort( 1,WGR[1] | 0x0100);  // Sets TRIG_POL for Falling Edge(inverts TRIG_IN)
              }
            else WritePort( 1,WGR[1] & 0xFEFF);  // Sets TRIG_POL for Rising Edge
        }
        else WritePort( 0,WGR[0] & 0x7FFF);  // Shuts down TRIG_EN

        WritePort( 1,WGR[1] | 0x1000);  // Enables VCO
// if (ClockRate < 60000000) SetDDS(ClockRate);  // Enables DDS if selected

        WritePort(1,WGR[1] & 0xFFFE);    // Enables MAD Counter.

     }

//WritePort(1,(WGR[1] & 0x7FFF));
}

//---------------------------------------------------------------------------

void da11000_SetSoftTrigger(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if ( (WGR[1] & 0x0100) > 0 ) {
        WritePort( 1,WGR[1] & 0xFEFF);
        WritePort( 1,WGR[1] | 0x0100);
    }
    else {
        WritePort( 1,WGR[1] | 0x0100);
        WritePort( 1,WGR[1] & 0xFEFF);
    }
}

//---------------------------------------------------------------------------

void da11000_SetMarkers(DWORD CardNum, DWORD PointAddr, BYTE Nib1, BYTE Nib2)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    WORD ReadBack;
    PointAddr = PointAddr*2; // Scale up to account for byte/word conversion

    DWORD Nib1x,Nib2x;     // Change to DWORD from BYTE
    Nib1x = Nib1;
    Nib2x = Nib2;
    Nib1x = Nib1x << 12;  // Move Nib1 over to upper nibble.
    Nib2x = Nib2x << 12;  // Move Nib2 over to upper nibble.

// Place marker Nib1x at address zero (0x16) PLUS MODULO 16d (Use rounding)

    DWORD NewAddrNib1 = (PointAddr & 0xFFFFFFE0) | 0x6;   // Actually 0110 => 011 or 3
    ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib1);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib1, ReadBack | Nib1x);

    NewAddrNib1 = (PointAddr & 0xFFFFFFE0) | 0x16;
    ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib1);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib1, ReadBack | Nib1x);

// Place marker Nib2 at address zero (0x16) PLUS MODULO 16d (Use rounding)

    DWORD NewAddrNib2 = (PointAddr & 0xFFFFFFE0) | 0x4;   // Actually 0110 => 010 or 2
    ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib2);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib2, ReadBack | Nib2x);

    NewAddrNib2 = (PointAddr & 0xFFFFFFE0) | 0x14;
    ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib2);
    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, NewAddrNib2, ReadBack | Nib2x);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

}

//---------------------------------------------------------------------------

void da11000_CreateSingleSegment(DWORD CardNum, DWORD ChanNum, DWORD NumPoints, DWORD NumLoops, PVOID UserArrayPtr, DWORD TrigEn)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

//   clock_t start, end; double elapsed;
//   start = clock();


// Put PCI local bus to R/W mode
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);
//WritePort(1,(WGR[1] | 0x8000));

    DWORD StartAddr = 0;                // Always starts at address zero.

    NumPoints = NumPoints & 0xFFFFFFC0; // Make sure number is "MOD 4x16=64"
                                        // -> This is minimum segment size
                                        //    and increment size.

    DWORD PipeLineOffset = 64;  // Add PipeLine Purge offset
    DWORD PadValue = 0;         // Use default value for loop

// USE BACK-END PADDING IF EXTERNAL TRIGGER USED
    if (TrigEn) {
        PadValue = 64;                                                              //  _|_
        for (DWORD i=PipeLineOffset+NumPoints-1; i < PipeLineOffset+NumPoints-1+PadValue+1; i++) {     // Fill another 64 samples
           P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (i)*2, 2047 /*1000*/);
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// WRITE USER DATA
//    P9030_WriteSpaceBlock(da11000_hPlx, (PipeLineOffset)*2, UserArrayPtr, (NumPoints*2),
//                          P9030_ADDR_SPACE0, P9030_MODE_WORD);
//

//  FILL BEGINNING PIPELINE
    for (DWORD i=0; i < PipeLineOffset; i++) {
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (i)*2, 2047 /*1000*/);
    }

//  FILL DATA
    DWORD ArrayOffset = 0;
    for (DWORD i=0; i < NumPoints; i=i+16) {
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +11)*2, *((WORD *)UserArrayPtr + ArrayOffset +  0) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +10)*2, *((WORD *)UserArrayPtr + ArrayOffset +  4) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 3)*2, *((WORD *)UserArrayPtr + ArrayOffset +  8) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 2)*2, *((WORD *)UserArrayPtr + ArrayOffset + 12) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 9)*2, *((WORD *)UserArrayPtr + ArrayOffset +  1) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 8)*2, *((WORD *)UserArrayPtr + ArrayOffset +  5) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 1)*2, *((WORD *)UserArrayPtr + ArrayOffset +  9) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 0)*2, *((WORD *)UserArrayPtr + ArrayOffset + 13) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +15)*2, *((WORD *)UserArrayPtr + ArrayOffset +  2) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +14)*2, *((WORD *)UserArrayPtr + ArrayOffset +  6) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 7)*2, *((WORD *)UserArrayPtr + ArrayOffset + 10) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 6)*2, *((WORD *)UserArrayPtr + ArrayOffset + 14) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +13)*2, *((WORD *)UserArrayPtr + ArrayOffset +  3) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i +12)*2, *((WORD *)UserArrayPtr + ArrayOffset +  7) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 5)*2, *((WORD *)UserArrayPtr + ArrayOffset + 11) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset + i + 4)*2, *((WORD *)UserArrayPtr + ArrayOffset + 15) );
    ArrayOffset = ArrayOffset + 16;

    }


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

NumPoints = NumPoints + PadValue;

// WRITE COMMAND DATA + REPEAT COUNT TO RAM CHIPS
    DWORD INPUT_CMD  = 3;
    DWORD CMD_OFFSET = 15;  //actual = 15

    DWORD INPUT_RPT_CNT  = 0;  // put back NumLoops;
    DWORD INPUT_JMP_ADDR = 0;  // Loop only.
    DWORD INPUT_BEG_ADDR = 0;

// WRITE INPUT_CMD
    WORD TempWord,TempCMD;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1)*2 );  // Read last WORD of segment
    TempCMD  =  INPUT_CMD << 14;
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1)*2, TempCMD);

    // Second Bin
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1-16)*2 );  // Read 2nd to last WORD of segment
    TempCMD  =  INPUT_CMD << 14;
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1-16)*2, TempCMD);

// WRITE INPUT_JMP_ADDR

    CMD_OFFSET = 6;  // FIX THIS  // actual = 6
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1)*2 );  // Read last WORD of segment
    if (TrigEn) TempCMD  =  0; else TempCMD  =  1 << 12;   // Jump Back to end of BegPad if NOT TRIGGERED
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1)*2, TempCMD);

    // Second Bin
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1-16)*2 );  // Read 2nd to last WORD of segment
    if (TrigEn) TempCMD  =  0; else TempCMD  =  1 << 12;   // Jump Back to end of BegPad if NOT TRIGGERED
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (PipeLineOffset-CMD_OFFSET+NumPoints-1-16)*2, TempCMD);


// SET MARKERS
//    da11000_SetMarkers(128, 0xFF, 0xFF);

// Put PCI local bus back to original state
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

//WritePort(1,(WGR[1] & 0x7FFF));

//end = clock();
//elapsed = ((double) (end - start)) / CLOCKS_PER_SEC;
//   lapse = (end-start)/CLK_TCK;

//  COMMAND SEQUENCING PROGRESSION (Minimum comand = 2 + 14 + 16 + 16 = 48 = CMD+RPT+JMP+BEG)
//
//  ADDR  X()        Comment
//  ----  ----       -------
//     0  X(quad-3)  Result of Previous Command if present
//     1  X(quad-2)  not used
//     2  X(quad-1)  Standard Command = <INPUT_CMD, INPUT_RPT_CNT, INPUT_JMP_ADDR, INPUT_BEG_ADDR>
//     3  X(quad)    Standard Command = <INPUT_CMD, INPUT_RPT_CNT, INPUT_JMP_ADDR, INPUT_BEG_ADDR>
//     ?  X(quad+1)  Result of Previous Command if present

//  COMMAND LIST
//
//  CMD   ACTION
//  ----  ------
//   00   no-op
//        <MAD_CNT increments normally on FAST_CNT>
//
//   01   IF (RPT_CNT == 0) THEN LOAD JMP ADDR  (if JMP=BEG, then loop forever)
//        IF (RPT_CNT != 0) THEN LOAD BEG ADDR
//
//   10   Stop All Counters. Start-up when Trigger Condition Met.
//
//   11   Stop All Counters. Stop VCO. Start-up when Trigger Condition Met.
//
//
}


//---------------------------------------------------------------------------

void WriteData(DWORD  StartAddr,
               DWORD  NumPts,
               DWORD  BeginPad,
               DWORD  EndingPad,
               PVOID SegmentPtr) {

    DWORD PipeLineOffset = 64;    // Leading Pad Value
    DWORD PadValue = 64;          // Back End Pad Value

    // WRITE BACK-END PADDING (64 WORDS)                                             V
        for (DWORD i=StartAddr+PipeLineOffset+NumPts-1; i < StartAddr+PipeLineOffset+NumPts-1+PadValue+1; i++) {     // Fill another 64 samples
            P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (i)*2, EndingPad);
        }
//    // WRITE USER DATA
//        P9030_WriteSpaceBlock(da11000_hPlx, (StartAddr+PipeLineOffset)*2, SegmentPtr, (NumPts*2),
//                              P9030_ADDR_SPACE0, P9030_MODE_WORD);


//  FILL DATA
    DWORD ArrayOffset = 0;
    for (DWORD i=0; i < NumPts; i=i+16) {
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +11)*2, *((WORD *)SegmentPtr + ArrayOffset +  0) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +10)*2, *((WORD *)SegmentPtr + ArrayOffset +  4) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 3)*2, *((WORD *)SegmentPtr + ArrayOffset +  8) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 2)*2, *((WORD *)SegmentPtr + ArrayOffset + 12) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 9)*2, *((WORD *)SegmentPtr + ArrayOffset +  1) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 8)*2, *((WORD *)SegmentPtr + ArrayOffset +  5) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 1)*2, *((WORD *)SegmentPtr + ArrayOffset +  9) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 0)*2, *((WORD *)SegmentPtr + ArrayOffset + 13) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +15)*2, *((WORD *)SegmentPtr + ArrayOffset +  2) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +14)*2, *((WORD *)SegmentPtr + ArrayOffset +  6) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 7)*2, *((WORD *)SegmentPtr + ArrayOffset + 10) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 6)*2, *((WORD *)SegmentPtr + ArrayOffset + 14) );

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +13)*2, *((WORD *)SegmentPtr + ArrayOffset +  3) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i +12)*2, *((WORD *)SegmentPtr + ArrayOffset +  7) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 5)*2, *((WORD *)SegmentPtr + ArrayOffset + 11) );
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr + PipeLineOffset + i + 4)*2, *((WORD *)SegmentPtr + ArrayOffset + 15) );
    ArrayOffset = ArrayOffset + 16;

    }

    // WRITE PRE-DATA PADDING  (64 WORDS)
        for (DWORD i=StartAddr; i < StartAddr+PipeLineOffset; i++) {
            P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (i)*2, BeginPad);
        }
}

void WriteCmd(DWORD  StartAddr,
              DWORD  NumPts,
              DWORD TrigInfo) {

    DWORD INPUT_CMD;
    DWORD CMD_OFFSET = 15;  // Keep in mind that actual system ram position is (15 - CMD_OFFSET)
    DWORD Trig_Offset;

    if (TrigInfo == 0) {
        INPUT_CMD = 1;
        Trig_Offset = 0;
    }
    else {
        INPUT_CMD = 3;
        Trig_Offset = 64;
    }

    DWORD PipeLineOffset;
    PipeLineOffset = 64 + Trig_Offset;    // Leading Pad Value + backend pad offset if trigger mode

   // WRITE INPUT_CMD
    WORD TempWord,TempCMD;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempCMD  =  INPUT_CMD << 14;
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCMD);

    // Second Bin
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempCMD  =  INPUT_CMD << 14;
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCMD);

    // Third Bin
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempCMD  =  INPUT_CMD << 14;
    TempCMD = TempCMD | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCMD);


// WRITE ADDITIONAL INPUT_CMD FOR TRIGGERED SEGMENT (CATCHES RPT COUNTS)
    if (TrigInfo != 0) {
        INPUT_CMD = 1;
        Trig_Offset = 0;

        PipeLineOffset = 64; // + Trig_Offset;    // Leading Pad Value + backend pad offset if trigger mode

        // First Bin
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
        TempCMD  =  INPUT_CMD << 14;
        TempCMD = TempCMD | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCMD);

        // Second Bin
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
        TempCMD  =  INPUT_CMD << 14;
        TempCMD = TempCMD | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCMD);

        // Third Bin
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
        TempCMD  =  INPUT_CMD << 14;
        TempCMD = TempCMD | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCMD);
    }


}

void WriteJmpAddr(DWORD  StartAddr,
                  DWORD  NumPts,
                  DWORD  JumpAddr,
                  DWORD  TrigInfo) {

    DWORD Trig_Offset;

    if (TrigInfo == 0) {
        Trig_Offset = 0;
        JumpAddr = JumpAddr + 64;  // Jump over 2 pads
    }
    else {
        Trig_Offset = 64;
    }

    DWORD PipeLineOffset = 64 + Trig_Offset;    // Leading Pad Value

    // WRITE INPUT_JMP_ADDR
    // First Bin
    WORD TempJmp;
    DWORD Input_Jmp_Shifted = JumpAddr >> 6;  // Compensates for 64 byte increments for Jmp Loading

    DWORD CMD_OFFSET = 6;
    DWORD TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
    CMD_OFFSET = 7;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
    CMD_OFFSET = 8;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
    CMD_OFFSET = 9;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);

    // Second Bin
    CMD_OFFSET = 6;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
    CMD_OFFSET = 7;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
    CMD_OFFSET = 8;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
    CMD_OFFSET = 9;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);

    // Third Bin
    CMD_OFFSET = 6;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
    CMD_OFFSET = 7;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
    CMD_OFFSET = 8;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
    CMD_OFFSET = 9;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
    TempJmp = TempJmp | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);


// WRITE ADDITIONAL INPUT_JMP_ADDR FOR TRIGGERED SEGMENT (CATCHES RPT COUNTS)
    if (TrigInfo != 0) {

        PipeLineOffset = 64; // + Trig_Offset;    // Leading Pad Value + backend pad offset if trigger mode

        // First Bin
        Input_Jmp_Shifted = (StartAddr+PipeLineOffset+NumPts) >> 6;  // Create mini jump to next memory location, let cmd at end of pad take care of real jump

        CMD_OFFSET = 6;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
        CMD_OFFSET = 7;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
        CMD_OFFSET = 8;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);
        CMD_OFFSET = 9;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempJmp);

        // Second Bin
        CMD_OFFSET = 6;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
        CMD_OFFSET = 7;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
        CMD_OFFSET = 8;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);
        CMD_OFFSET = 9;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempJmp);

        // Third Bin
        CMD_OFFSET = 6;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 12) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
        CMD_OFFSET = 7;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 8) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
        CMD_OFFSET = 8;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 4) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
        CMD_OFFSET = 9;
        TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
        TempJmp  =  (Input_Jmp_Shifted << 0) & 0xF000;
        TempJmp = TempJmp | TempWord;
        P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempJmp);
    }


}

void WriteBegAddr(DWORD  StartAddr,    // Also Beginning Address
                  DWORD  NumPts,
                  DWORD TrigInfo) {

    DWORD Trig_Offset;
    DWORD Input_Beg_Shifted;

    if (TrigInfo == 0) {
        Trig_Offset = 0;
        Input_Beg_Shifted = StartAddr + 64;  // Skips starting pad during loop
    }
    else {
        Trig_Offset = 64;
        Input_Beg_Shifted = StartAddr;  // Starts at beginning of pad for triggered looping
    }

    DWORD PipeLineOffset = 64 + Trig_Offset;    // Leading Pad Value

    // WRITE INPUT_BEG_ADDR
    // First Bin
    WORD TempBeg;
    Input_Beg_Shifted = Input_Beg_Shifted >> 6;  // Compensates for 64 byte increments

    DWORD CMD_OFFSET = 0;
    DWORD TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);

    // Second Bin
    CMD_OFFSET = 0;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);

    // Third Bin
    CMD_OFFSET = 0;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);


// WRITE  ADDITIONAL INPUT_BEG_ADDR FOR TRIGGERED SEGMENT (CATCHES RPT COUNTS)

    if (TrigInfo != 0) {
        Trig_Offset = 0;
        Input_Beg_Shifted = StartAddr + 64;  // Skips starting pad during loop

    PipeLineOffset = 64;// + Trig_Offset;    // Leading Pad Value

    // First Bin
    Input_Beg_Shifted = Input_Beg_Shifted >> 6;  // Compensates for 64 byte increments

    CMD_OFFSET = 0;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempBeg);

    // Second Bin
    CMD_OFFSET = 0;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempBeg);

    // Third Bin
    CMD_OFFSET = 0;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 12) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 1;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 8) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 2;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 4) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    CMD_OFFSET = 3;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempBeg  =  (Input_Beg_Shifted << 0) & 0xF000;
    TempBeg = TempBeg | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempBeg);
    }


}


void WriteRepCnt(DWORD  StartAddr,    // Also Beginning Address
                 DWORD  NumPts,
                 DWORD TrigInfo,
                 DWORD NumLoops) {

    DWORD Trig_Offset;

    if (TrigInfo == 0) {
        Trig_Offset = 0;
    }
    else {
        Trig_Offset = 64;
    }

    DWORD PipeLineOffset = 64 + Trig_Offset;    // Leading Pad Value

    // WRITE INPUT_RPT_CNT
    // First Bin
    WORD TempCnt;
    DWORD Repeat_Cnt = NumLoops;

    DWORD CMD_OFFSET = 10;
    DWORD TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 12) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCnt);
    CMD_OFFSET = 11;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 8) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCnt);
    CMD_OFFSET = 14;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 4) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCnt);
    CMD_OFFSET = 15;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 0) & 0x3000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1)*2, TempCnt);

    // Second Bin
    CMD_OFFSET = 10;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 12) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCnt);
    CMD_OFFSET = 11;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 8) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCnt);
    CMD_OFFSET = 14;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 4) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCnt);
    CMD_OFFSET = 15;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 0) & 0x3000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-16)*2, TempCnt);

    // Third Bin
    CMD_OFFSET = 10;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 12) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCnt);
    CMD_OFFSET = 11;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 8) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCnt);
    CMD_OFFSET = 14;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 4) & 0xF000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCnt);
    CMD_OFFSET = 15;
    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2 );  // Read last WORD of segment
    TempCnt  =  (Repeat_Cnt << 0) & 0x3000;
    TempCnt = TempCnt | TempWord;
    P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-CMD_OFFSET+NumPts-1-32)*2, TempCnt);

}


void da11000_CreateSegments(DWORD CardNum, DWORD ChanNum, DWORD NumSegments, PVOID PtrToSegmentsList)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

// Put PCI local bus to R/W mode
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

// SETUP DYNAMIC ARRAY TO SEGMENT POINTERS
// Multiple Segment Structure
    typedef struct                // Creates Type instead of variable
    {
        DWORD    SegmentNum;     // Current Segment Number
        PVOID    SegmentPtr;     // Pointer to current user segment
                                 // ==> elements of one diminsional array must be of type WORD
        DWORD    NumPoints;      // Number of points in segment
        DWORD    NumLoops;       // Number of times to repeat segment (applies to next segment)
        DWORD    BeginPadVal;    // Pad value for beginning of triggered segment
        DWORD    EndingPadVal;   // Pad value for ending of triggered segment
        DWORD    TrigEn;         // If > 0 then wait for trigger before going to next segment

        DWORD    NextSegNum;     // Next segment to jump to after completion
                                 // of current segment activities 
    } SegmentStruct;
//

    SegmentStruct *SegmentsList;  // Create Pointer to data of type SegmentStruct
    try { SegmentsList = new SegmentStruct[NumSegments]; }
    catch (std::bad_alloc) {exit(-1);}

    PVOID BackupPointer = SegmentsList;
    SegmentsList = (SegmentStruct*) PtrToSegmentsList;   // Assign user pointer to PSegmentsList

///////////////////////////////////////////////
// *****  START SEGMENT CREATION LOOPS ***** //
///////////////////////////////////////////////

    DWORD Running_StartAddr = 0;
    DWORD SegmentStartAddr[32768];  // Store start addresses for each segment for jump reference

// Write Data + pads
    for (WORD i=0; i < NumSegments; i++) {
        SegmentStartAddr[i] = Running_StartAddr;  // Load start addresses for

        WriteData(Running_StartAddr,            // DWORD  StartAddr,
                  SegmentsList[i].NumPoints,    // DWORD  NumPts,
                  SegmentsList[i].BeginPadVal,  // DWORD  BeginPad,
                  SegmentsList[i].EndingPadVal, // DWORD  EndingPad,
                  SegmentsList[i].SegmentPtr    // DPVOID SegmentPtr
                  );
        Running_StartAddr = Running_StartAddr + 128 + SegmentsList[i].NumPoints;    // Prepare for next segment
    }

// Write Commands (00,01,10,11)
    for (WORD i=0; i < NumSegments; i++) {
        WriteCmd(SegmentStartAddr[i],               // DWORD  StartAddr,
                 SegmentsList[i].NumPoints,         // DWORD  NumPts,
                 SegmentsList[i].TrigEn             // DWORD TrigInfo
                 );
    }

// Write Jump Addresses
   for (WORD i=0; i < NumSegments; i++) {
       WriteJmpAddr(SegmentStartAddr[i],                           // DWORD  StartAddr,
                    SegmentsList[i].NumPoints,                     // DWORD  NumPts,
                    SegmentStartAddr[SegmentsList[i].NextSegNum],  // DWORD  JumpAddr
                    SegmentsList[i].TrigEn                         // DWORD TrigInfo
                    );
       WriteBegAddr(SegmentStartAddr[i],             // DWORD  StartAddr,
                    SegmentsList[i].NumPoints,       // DWORD  NumPts,
                    SegmentsList[i].TrigEn           // DWORD TrigInfo
                    );
       WriteRepCnt(SegmentStartAddr[i],              // DWORD  StartAddr,
                   SegmentsList[i].NumPoints,        // DWORD  NumPts,
                   SegmentsList[i].TrigEn,           // DWORD TrigInfo
                   SegmentsList[i].NumLoops          // DWORD NumLoops
                   );
       }
// Write Repeat Count


//////////////////////////////////////////////
// *****  END SEGMENT CREATION LOOPS  ***** //
//////////////////////////////////////////////


    SegmentsList = (SegmentStruct*) BackupPointer;
    delete[] SegmentsList;       // Free memory

// Put PCI local bus back to original state
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
}

//---------------------------------------------------------------------------

void ClearCodes(DWORD  StartAddr,
                DWORD  NumPts) {

    DWORD INPUT_CMD;
    DWORD CMD_OFFSET = 15;  // Keep in mind that actual system ram position is (15 - CMD_OFFSET)
    DWORD PipeLineOffset;
    WORD TempWord,TempCMD;

// Delete non-triggered codes
    PipeLineOffset = 64;
    for (WORD i=0; i < 32; i++) {
      TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-i+NumPts-1)*2 );  // Read last WORD of segment
      P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-i+NumPts-1)*2, (0x0FFF & TempWord));  // erases code words
    }
    PipeLineOffset = 128;   // This covers triggered mode codes
    for (WORD i=0; i < 32; i++) {
      TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-i+NumPts-1)*2 );  // Read last WORD of segment
      P9030_WriteSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, (StartAddr+PipeLineOffset-i+NumPts-1)*2, (0x0FFF & TempWord));  // erases code words
    }

}


void da11000_UpdateSegmentCmds(DWORD CardNum, DWORD ChanNum, DWORD NumSegments, PVOID PtrToSegmentsList)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

// Put PCI local bus to R/W mode
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

// SETUP DYNAMIC ARRAY TO SEGMENT POINTERS
// Multiple Segment Structure
    typedef struct                // Creates Type instead of variable
    {
        DWORD    SegmentNum;     // Current Segment Number
        PVOID    SegmentPtr;     // Pointer to current user segment
                                 // ==> elements of one diminsional array must be of type WORD
        DWORD    NumPoints;      // Number of points in segment
        DWORD    NumLoops;       // Number of times to repeat segment (applies to next segment)
        DWORD    BeginPadVal;    // Pad value for beginning of triggered segment
        DWORD    EndingPadVal;   // Pad value for ending of triggered segment
        DWORD    TrigEn;         // If > 0 then wait for trigger before going to next segment

        DWORD    NextSegNum;     // Next segment to jump to after completion
                                 // of current segment activities
    } SegmentStruct;
//

    SegmentStruct *SegmentsList;  // Create Pointer to data of type SegmentStruct
    try { SegmentsList = new SegmentStruct[NumSegments]; }
    catch (std::bad_alloc) {exit(-1);}

    PVOID BackupPointer = SegmentsList;
    SegmentsList = (SegmentStruct*) PtrToSegmentsList;   // Assign user pointer to PSegmentsList

///////////////////////////////////////////////
// *****  START SEGMENT CREATION LOOPS ***** //
///////////////////////////////////////////////

    DWORD Running_StartAddr = 0;
    DWORD SegmentStartAddr[32768];  // Store start addresses for each segment for jump reference

// Zero out code areas
    for (WORD i=0; i < NumSegments; i++) {
        SegmentStartAddr[i] = Running_StartAddr;  // Load start addresses for

        ClearCodes(SegmentStartAddr[i],           // DWORD  StartAddr,
                   SegmentsList[i].NumPoints      // DWORD  NumPts
                   );
        Running_StartAddr = Running_StartAddr + 128 + SegmentsList[i].NumPoints;    // Prepare for next segment
    }

// Write Commands (00,01,10,11)
    for (WORD i=0; i < NumSegments; i++) {
        WriteCmd(SegmentStartAddr[i],               // DWORD  StartAddr,
                 SegmentsList[i].NumPoints,         // DWORD  NumPts,
                 SegmentsList[i].TrigEn             // DWORD TrigInfo
                 );
    }

// Write Jump Addresses
   for (WORD i=0; i < NumSegments; i++) {
       WriteJmpAddr(SegmentStartAddr[i],                           // DWORD  StartAddr,
                    SegmentsList[i].NumPoints,                     // DWORD  NumPts,
                    SegmentStartAddr[SegmentsList[i].NextSegNum],  // DWORD  JumpAddr
                    SegmentsList[i].TrigEn                         // DWORD TrigInfo
                    );
       WriteBegAddr(SegmentStartAddr[i],             // DWORD  StartAddr,
                    SegmentsList[i].NumPoints,       // DWORD  NumPts,
                    SegmentsList[i].TrigEn           // DWORD TrigInfo
                    );
       WriteRepCnt(SegmentStartAddr[i],              // DWORD  StartAddr,
                   SegmentsList[i].NumPoints,        // DWORD  NumPts,
                   SegmentsList[i].TrigEn,           // DWORD TrigInfo
                   SegmentsList[i].NumLoops          // DWORD NumLoops
                   );
       }


//////////////////////////////////////////////
// *****  END SEGMENT CREATION LOOPS  ***** //
//////////////////////////////////////////////


    SegmentsList = (SegmentStruct*) BackupPointer;
    delete[] SegmentsList;       // Free memory

// Put PCI local bus back to original state
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
}


//---------------------------------------------------------------------------

#ifndef linux
void da11000_WritePoint(DWORD CardNum, BYTE BufferNum, DWORD PointAddr, WORD PointValue)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,0, 0);

    P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, PointAddr*2, PointValue);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,0, 0);
}

//---------------------------------------------------------------------------

WORD da11000_ReadPoint(DWORD CardNum, BYTE BufferNum, DWORD PointAddr)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,0, 0);

    WORD TempWord;

    TempWord = P9030_ReadSpaceWord(da11000_hPlx, P9030_ADDR_SPACE0, PointAddr*2);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,1, 0);

    return(TempWord);
}

//---------------------------------------------------------------------------

void da11000_WriteBlock(DWORD CardNum, BYTE BufferNum, DWORD StartAddr, DWORD Range, PVOID UserArrayPtr )
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,0, 0);     // ATTEN_hPlx   da11000_hPlx

    P9030_WriteSpaceBlock(da11000_hPlx, StartAddr, UserArrayPtr, Range*2,
                          P9030_ADDR_SPACE0, P9030_MODE_WORD);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
//    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,1, 0);
}

//---------------------------------------------------------------------------

//void da11000_ReadBlock(DWORD CardNum, BYTE BufferNum, DWORD StartAddr, DWORD Range, PVOID UserArrayPtr )
//{
//    if (!BrdNumOK(CardNum)) return;
//    da11000_hPlx = GetBrdHandle(CardNum);
//
//    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);
////    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,0, 0);
//
//    P9030_ReadSpaceBlock(da11000_hPlx, StartAddr, UserArrayPtr, Range*2,
//                          P9030_ADDR_SPACE0, P9030_MODE_WORD);
//
//    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);
////    if (PCI_SEL_MODE > 0) PCI_Bus_Enable(CardNum,1, 0);
//}
//

//***************************************************************************

DWORD da11000_ReadPCIConfigReg(DWORD BrdNum, DWORD dwReg)
{
    if (!BrdNumOK(BrdNum)) return(0);
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    DWORD RegValue = 0;

    RegValue = P9030_ReadPCIReg(da11000_hPlx, dwReg);
    return(RegValue);
}

void da11000_WritePCIConfigReg(DWORD BrdNum, DWORD dwReg, DWORD RegValue)
{
    if (!BrdNumOK(BrdNum)) return;
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    P9030_WriteReg (da11000_hPlx, dwReg, RegValue);
    return;
}

//***************************************************************************

DWORD da11000_ReadPCILocalReg(DWORD BrdNum, DWORD dwReg)
{
    if (!BrdNumOK(BrdNum)) return(0);
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    DWORD RegValue = 0;

    RegValue = P9030_ReadReg(da11000_hPlx, dwReg);  // Same => P9030_ReadSpaceDWord(AD4200_1_hPlx, P9030_ADDR_REG, dwReg);
    return(RegValue);

}

void da11000_WritePCILocalReg(DWORD BrdNum, DWORD dwReg, DWORD RegValue)
{
    if (!BrdNumOK(BrdNum)) return;
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    P9030_WriteReg(da11000_hPlx, dwReg, RegValue);
    return;

}

//***************************************************************************

/*
DWORD da11000_EEPROMReadDWord(DWORD CardNum, DWORD dwReg)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    DWORD RegValue = 0;
    P9030_EEPROMReadDWord(da11000_hPlx, dwReg, &RegValue);

    return(RegValue);
}

void da11000_EEPROMWriteDWord(DWORD CardNum, DWORD dwReg, DWORD RegValue)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    P9030_EEPROMWriteDWord(da11000_hPlx, dwReg, RegValue);

    return;
}
*/

/*****************************************************************************/
/*****************************************************************************/
#endif

DWORD P9030_CNTRL_DATA;

void EEPROM_DELAY(DWORD BrdNum)
{
    if (!BrdNumOK(BrdNum)) return;
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

//    clock_t start, end;
//    start = clock();

    for (int x = 0; x < 500; x++)   // Was "33; x++)". Changed 3/10/05
    P9030_CNTRL_DATA = P9030_ReadReg(da11000_hPlx, P9030_CNTRL);  // Save 0x54 Register Value

//    end = clock();
//    printf("The time was: %f\n", (end - start) / CLK_TCK);

    return;
}

#ifndef linux
/*****************************************************************************/

void WriteSer(DWORD BrdNum, BYTE BinValue)
{
    if (!BrdNumOK(BrdNum)) return;
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

//DWORD P9030_ReadReg(P9030_HANDLE hPlx, DWORD dwReg);
//void P9030_WriteReg(P9030_HANDLE hPlx, DWORD dwReg, DWORD dwData);

// EECS, EESK, EEDI, EEDO

EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0
EEPROM_DELAY(BrdNum);

    if (BinValue == 1) {
        P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x04000000);  // Set SER_DIN to 1
    }

EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x01000000);  // Set SER_CLK to 1
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Resets both SER_DIN, SER_CLK
EEPROM_DELAY(BrdNum);

}

/*****************************************************************************/

DWORD da11000_EEPROMReadDWord(DWORD BrdNum, DWORD Addr)
{
    if (!BrdNumOK(BrdNum)) return(0);
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    Addr = Addr / 2;  // Normalize to word boundaries

// EXIT IF PARAMETERS OUT OF RANGE
    if (Addr > 127) return(0);

    P9030_CNTRL_DATA = P9030_ReadReg(da11000_hPlx, P9030_CNTRL);  // Save 0x54 Register Value

// ---------WRITE ENABLE---------------------------------------------
// Make sure SER_DIN, SER_CLK are low
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0
EEPROM_DELAY(BrdNum);

// SET "93C56" CS HIGH TO ENABLE CLOCKING ACTION
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);
EEPROM_DELAY(BrdNum);

// LOAD WEN COMMAND FIRST
    WORD WEN_Command = 0x04C0;
    for (int x = 0; x < 11; x++) {
        if ( ( WEN_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        WEN_Command = WEN_Command << 1;
    }
EEPROM_DELAY(BrdNum);
// SET "93C56" CS LOW TO FINISH WRITE
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);
EEPROM_DELAY(BrdNum);

// ---------READ FIRST WORD---------------------------------------------

// SET "CS_93C56" HIGH TO ENABLE CLOCKING ACTION
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);
EEPROM_DELAY(BrdNum);

// LOAD READ COMMAND + ADDRESS
    WORD READ_Command = 0x0600 | Addr;
    for (int x = 0; x < 11; x++) {
        if ( ( READ_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        READ_Command = READ_Command << 1;
    }
EEPROM_DELAY(BrdNum);

// READ ACTUAL E2PROM DATA
    DWORD READ_Data = 0;
    WriteSer(BrdNum, 0);  // Send 1st clock for dummy "0"
EEPROM_DELAY(BrdNum);
    for (int x = 0; x < 32; x++) {
        DWORD TestWord = P9030_ReadReg(da11000_hPlx, P9030_CNTRL);
EEPROM_DELAY(BrdNum);
        READ_Data = READ_Data << 1;
        if ( (TestWord & 0x08000000) > 0 )
            READ_Data = READ_Data + 1;
        WriteSer(BrdNum, 0); // Send clock to read next bit
    }
EEPROM_DELAY(BrdNum);

// SET "CS_93C56" LOW TO FINISH WRITE
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);

    return(READ_Data);
}

/*****************************************************************************/

void da11000_EEPROMWriteDWord(DWORD BrdNum, DWORD Addr, DWORD RegValue)
{
    if (!BrdNumOK(BrdNum)) return;
    P9030_HANDLE da11000_hPlx = GetBrdHandle(BrdNum);

    Addr = Addr / 2;  // Normalize to word boundaries
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_PROT_AREA, 0x00000000);  // Removed Protected Area

// EXIT IF PARAMETERS OUT OF RANGE
    if (Addr > 128) return;

EEPROM_DELAY(BrdNum);
    P9030_CNTRL_DATA = P9030_ReadReg(da11000_hPlx, P9030_CNTRL);  // Save 0x54 Register Value

// ---------WRITE ENABLE---------------------------------------------
// Make sure SER_DIN, SER_CLK are low
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0

// SET "93C56" CS HIGH TO ENABLE CLOCKING ACTION
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);

// LOAD WEN COMMAND FIRST
    WORD WEN_Command = 0x04C0;
    for (int x = 0; x < 11; x++) {
        if ( ( WEN_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        WEN_Command = WEN_Command << 1;
    }
// SET "93C56" CS LOW TO FINISH WRITE
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);

    WORD WRITE_Command = 0x0500 | (Addr);

// --------WRITE FIRST WORD----------------------------------------------

// Make sure SER_DIN, SER_CLK are low
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0

// SET "CS_93C56" HIGH TO ENABLE CLOCKING ACTION
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);

// LOAD WRITE COMMAND + ADDRESS + DATA
    WRITE_Command = 0x0500 | (Addr);
    for (int x = 0; x < 11; x++) {
        if ( ( WRITE_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        WRITE_Command = WRITE_Command << 1;
    }
    for (int x = 0; x < 16; x++) {
        if ( ( RegValue & 0x80000000) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        RegValue = RegValue << 1;
    }

// SET "CS_93C56" LOW TO FINISH WRITE
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);

// ---------WRITE SECOND WORD---------------------------------------------

for (int x = 0; x < 500; x++) EEPROM_DELAY(BrdNum);

// Make sure SER_DIN, SER_CLK are low
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0

// SET "CS_93C56" HIGH TO ENABLE CLOCKING ACTION
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);

// LOAD WRITE COMMAND + ADDRESS + DATA    (Second Word)
    WRITE_Command = 0x0500 | (Addr+1);
    for (int x = 0; x < 11; x++) {
        if ( ( WRITE_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        WRITE_Command = WRITE_Command << 1;
    }
    for (int x = 0; x < 16; x++) {
        if ( ( RegValue & 0x80000000) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        RegValue = RegValue << 1;
    }

// SET "CS_93C56" LOW TO FINISH WRITE
EEPROM_DELAY(BrdNum);
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);


//    clock_t start, end;
//    start = clock();

for (int x = 0; x < 500; x++) EEPROM_DELAY(BrdNum);

//    end = clock();

/*
// ---------WRITE DISABLE---------------------------------------------

// Make sure SER_DIN, SER_CLK are low
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFAFFFFFF);  // Set SER_DIN, SER_CLK to 0

// SET "93C56" CS HIGH TO ENABLE CLOCKING ACTION
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA | 0x02000000);

// LOAD WDS COMMAND TO PREVENT FURTHER WRITES
    WEN_Command = 0x0400;
    for (int x = 0; x < 11; x++) {
        if ( ( WEN_Command & 0x0400) > 0 ) WriteSer(BrdNum, 1);
        else WriteSer(BrdNum, 0);
        WEN_Command = WEN_Command << 1;
    }
// SET "93C56" CS LOW TO FINISH WRITE
    P9030_WriteReg(da11000_hPlx, P9030_CNTRL, P9030_CNTRL_DATA & 0xFDFFFFFF);
*/

}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

void RunTest(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    for(int i=0; i<1000000; i++) {
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W0, (WGR[0] & 0xFFFE));
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE1, W0, (WGR[0] | 0x0001));
    }
}

//---------------------------------------------------------------------------

int MemTest(DWORD CardNum)
{   // void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
                     //
    for (DWORD i=0; i < TestRange; i+=2) {
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != 0)
        { Errors = Errors + 1; }

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0xFFFF);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != 0xFFFF)
        { Errors = Errors + 1; }
    }


    for (DWORD i=0; i < TestRange; i+=2) {
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, (i & 0x0000FFFF) );
    }
    for (DWORD i=0; i < TestRange; i+=2) {
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != (i & 0xFFFF))
        { Errors = Errors + 1; }
    }


// USE RANDOM DATA TO TEST FOR MEMORY OVERLAP ERRORS

    srand(1); WORD JunkWord;
    for (DWORD i=0; i < TestRange; i+=2) {
        JunkWord = random(0x10000);
        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, JunkWord );
    }
    srand(1);
    for (DWORD i=0; i < TestRange; i+=2) {
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        JunkWord = random(0x10000);
        if (ReadBack != JunkWord)
        { Errors = Errors + 1; }
    }


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int Sine_64x(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD DAC_Value = 2047;
    double pi = 3.14159265358979;

    for (DWORD i=0; i < TestRange; i+=2) {

        DAC_Value = ceil(2047 + 2047*cos(2*pi*i/128));

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, DAC_Value);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != DAC_Value)
        { Errors = Errors + 1; }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int Sine_32x(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD DAC_Value = 2047;
    double pi = 3.14159265358979;

    for (DWORD i=0; i < TestRange; i+=2) {

        DAC_Value = ceil(2047 + 2047*cos(2*pi*i/64));

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, DAC_Value);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != DAC_Value)
        { Errors = Errors + 1; }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int Sine_16x(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD DAC_Value = 2047;
    double pi = 3.14159265358979;

    for (DWORD i=0; i < TestRange; i+=2) {

        DAC_Value = ceil(2047 + 2047*cos(2*pi*i/32));

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, DAC_Value);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != DAC_Value)
        { Errors = Errors + 1; }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int Sine_8x(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD DAC_Value = 2047;
    double pi = 3.14159265358979;
    DWORD idac;

    for (DWORD i=0; i < TestRange; i+=2) {

//      DAC_Value = _lrotl(i & 0x1E,7);
//      DAC_Value = _lrotl(i & 0x3E,6);
//      DAC_Value = _lrotl(i & 0x7E,5);

      DAC_Value = ceil(2047 + 2047*cos(2*pi*i/16));

/*
    if ((i & 0x1E) == _lrotl( 0,1)) {idac = (i & 0x1FFFE0) + _lrotl( 8,1);}
    if ((i & 0x1E) == _lrotl( 1,1)) {idac = (i & 0x1FFFE0) + _lrotl( 9,1);}
    if ((i & 0x1E) == _lrotl( 2,1)) {idac = (i & 0x1FFFE0) + _lrotl(10,1);}
    if ((i & 0x1E) == _lrotl( 3,1)) {idac = (i & 0x1FFFE0) + _lrotl(11,1);}
    if ((i & 0x1E) == _lrotl( 4,1)) {idac = (i & 0x1FFFE0) + _lrotl(12,1);}
    if ((i & 0x1E) == _lrotl( 5,1)) {idac = (i & 0x1FFFE0) + _lrotl(13,1);}
    if ((i & 0x1E) == _lrotl( 6,1)) {idac = (i & 0x1FFFE0) + _lrotl(14,1);}
    if ((i & 0x1E) == _lrotl( 7,1)) {idac = (i & 0x1FFFE0) + _lrotl(15,1);}
    if ((i & 0x1E) == _lrotl( 8,1)) {idac = (i & 0x1FFFE0) + _lrotl( 0,1);}
    if ((i & 0x1E) == _lrotl( 9,1)) {idac = (i & 0x1FFFE0) + _lrotl( 1,1);}
    if ((i & 0x1E) == _lrotl(10,1)) {idac = (i & 0x1FFFE0) + _lrotl( 2,1);}
    if ((i & 0x1E) == _lrotl(11,1)) {idac = (i & 0x1FFFE0) + _lrotl( 3,1);}
    if ((i & 0x1E) == _lrotl(12,1)) {idac = (i & 0x1FFFE0) + _lrotl( 4,1);}
    if ((i & 0x1E) == _lrotl(13,1)) {idac = (i & 0x1FFFE0) + _lrotl( 5,1);}
    if ((i & 0x1E) == _lrotl(14,1)) {idac = (i & 0x1FFFE0) + _lrotl( 6,1);}
    if ((i & 0x1E) == _lrotl(15,1)) {idac = (i & 0x1FFFE0) + _lrotl( 7,1);}
*/

        idac = i;

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, idac, DAC_Value);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, idac);
        if (ReadBack != DAC_Value)
        { Errors = Errors + 1; }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int Sine_4x(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD DAC_Value = 2047;
    double pi = 3.14159265358979;

    for (DWORD i=0; i < TestRange; i+=2) {

        DAC_Value = ceil(2047 + 2047*cos(2*pi*i/8));

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, DAC_Value);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != DAC_Value)
        { Errors = Errors + 1; }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);

}

//---------------------------------------------------------------------------

int PlacePoint(DWORD CardNum)
{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

// void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD Mask5 = 0x1F;
                     //
    for (DWORD i=0; i < TestRange; i+=2) {
        if ( ((i & Mask5) == 0x0) | ((i & Mask5) == 0x10)) {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0xFFFF);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0xFFFF)
            { Errors = Errors + 1; }
        }
        else {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0x0000);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0x0000)
            { Errors = Errors + 1; }
        }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);
}

//---------------------------------------------------------------------------

int PlacePoint2(DWORD CardNum) // {return(0);}

{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

// void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD Mask2 = 0x3;
                     //
    for (DWORD i=0; i < TestRange; i+=2) {
        if ((i & Mask2) == 0x2) {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0xFFF);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0xFFF)
            { Errors = Errors + 1; }
        }
        else {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0x0);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0x0)
            { Errors = Errors + 1; }
        }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);
}

//---------------------------------------------------------------------------

int PlacePoint3(DWORD CardNum) // {return(0);}

{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

// void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );

    for (DWORD i=0; i < TestRange; i+=2) {
        if ((i & BIT4) == 0) {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0xFFFF);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0xFFFF)
            { Errors = Errors + 1; }
        }
        else {
            P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0x0);
            ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
            if (ReadBack != 0x0)
            { Errors = Errors + 1; }
        }
    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);
}

//---------------------------------------------------------------------------

int PlacePoint4(DWORD CardNum) // {return(0);}

{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

// void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD Mask13 = 0x1FFF;


    DWORD PosWord;
                     //
/*    for (DWORD i=0; i < TestRange; i+=2) {
        PosWord = (i & Mask5)/2;

        if ((i & Mask5) == 0) { PosWord = PosWord | 0x20;}

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, PosWord);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != PosWord)
        { Errors = Errors + 1; }

    }
*/
    for (DWORD i=0; i < TestRange; i+=2) {
        PosWord = (i & Mask13)/2;  // Same as shift right by 1

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, PosWord);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != PosWord)
        { Errors = Errors + 1; }

    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);
}


//---------------------------------------------------------------------------

int Zeros(DWORD CardNum) // {return(0);}

{
    if (!BrdNumOK(CardNum)) return(0);
    da11000_hPlx = GetBrdHandle(CardNum);

// void P9030_WriteSpaceWord (P9030HANDLE hPlx, P9030_ADDR addrSpace, DWORD dwOffset, WORD data);
    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,1, 0);

    int Errors = 0;
    WORD ReadBack;
    DWORD TestRange = ( 0x1FFFFF * MEMORY_OPTION );
    DWORD Mask13 = 0x1FFF;


    DWORD PosWord;

    for (DWORD i=0; i < TestRange; i+=2) {

        P9030_WriteSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i, 0x7FF);
        ReadBack = P9030_ReadSpaceWord (da11000_hPlx, P9030_ADDR_SPACE0, i);
        if (ReadBack != 0x7FF)
        { Errors = Errors + 1; }

    }

// Place marker at address zero

    da11000_SetMarkers(CardNum, 0, 0xFF, 0xFF);


    if (PCI_SEL_MODE == 0) PCI_Bus_Enable(CardNum,0, 0);

    return(Errors);
}

//---------------------------------------------------------------------------
#endif

void PCI_Bus_Enable(DWORD CardNum,BYTE PCI_SEL, BYTE Update)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    if (PCI_SEL == 0)
    {
        if (Update == 1) PCI_SEL_MODE = 0;
        WritePort(0,(WGR[0] & 0xBFFF));  // Deactivates PCI_SEL
//        for (DWORD x = 0; x < 0xFFFFFF; x++) {}  // Delays by ???
    }
    else
    {
        if (Update == 1) PCI_SEL_MODE = 1;
        WritePort(0,(WGR[0] | 0x4000));  // Activates PCI_SEL
    }

}

//---------------------------------------------------------------------------

void Set_RAM_Offset(DWORD CardNum, WORD RAM_Offset)
{
    if (!BrdNumOK(CardNum)) return;
    da11000_hPlx = GetBrdHandle(CardNum);

    RAM_Offset = RAM_Offset << 9;
    WritePort(1,(WGR[1] & 0xF1FF));       // Clear RAM_Offset bits
    WritePort(1,(WGR[1] | RAM_Offset));   // Apply RAM_Offset
}

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

#ifndef linux
bool NoiseBd_Detected(P9030HANDLE TEST_hPLX)
{
    if (TEST_hPLX == NULL) return(false);

    bool SPACE0_ACTIVE, SPACE1_ACTIVE;
    SPACE0_ACTIVE = P9030_IsAddrSpaceActive(TEST_hPLX, P9030_ADDR_SPACE0);
    SPACE1_ACTIVE = P9030_IsAddrSpaceActive(TEST_hPLX, P9030_ADDR_SPACE1);

    if ( SPACE0_ACTIVE && !SPACE1_ACTIVE ) return(true);
    else return(false);
}

//---------------------------------------------------------------------------
// Atten_dB: 0 <= Atten_dB <= 63.75  [for Weinschel 3200-2 attenuator]

void da11000_SetNoiseAtten(DOUBLE Atten_dB)
{
    if (ATTEN_hPlx == NULL) return;  // Continue only if Noise Card Present

    if (da11000_hPlx != NULL) P9030_Close(da11000_hPlx);    // Close da11000 if da11000 Present

    if (da11000_hPlx != NULL)
        P9030_Open(&ATTEN_hPlx, VendorID, DeviceID, WORD ((Arb_Location+1)&0x1));  // Open Noise Card
    if (da11000_hPlx == NULL)
        P9030_Open(&ATTEN_hPlx, VendorID, DeviceID, 0);  // Open Noise Card

    if ( (Atten_dB < 0) || (Atten_dB > 63.75) ) return;

    WORD BinAttenValue = ceil ((63.75 - Atten_dB)/0.25);
    P9030_WriteSpaceWord (ATTEN_hPlx, P9030_ADDR_SPACE0, 4, BinAttenValue);


    P9030_Close(ATTEN_hPlx);     // Close Noise Card    // Open da11000 if da11000 Present
    if (da11000_hPlx != NULL) P9030_Open(&da11000_hPlx, VendorID, DeviceID, WORD (Arb_Location & 0x1));
    return;
}

//---------------------------------------------------------------------------

void da11000_SetMainAtten(DOUBLE Atten_dB)
{
    if (ATTEN_hPlx == NULL) return;  // Continue only if Noise Card Present

    if (da11000_hPlx != NULL) P9030_Close(da11000_hPlx);    // Close da11000 if da11000 Present

    if (da11000_hPlx != NULL) P9030_Open(&ATTEN_hPlx, VendorID, DeviceID, WORD ((Arb_Location+1)&0x1));  // Open Noise Card
    if (da11000_hPlx == NULL) P9030_Open(&ATTEN_hPlx, VendorID, DeviceID, 0);  // Open Noise Card

    if ( (Atten_dB < 0) || (Atten_dB > 63.75) ) return;

    WORD BinAttenValue = ceil ((63.75 - Atten_dB)/0.25);
    P9030_WriteSpaceWord (ATTEN_hPlx, P9030_ADDR_SPACE0, 2, BinAttenValue);

    P9030_Close(ATTEN_hPlx);     // Close Noise Card    // Open da11000 if da11000 Present
    if (da11000_hPlx != NULL) P9030_Open(&da11000_hPlx, VendorID, DeviceID, WORD (Arb_Location & 0x1));
    return;
}
#endif


//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

void da11000_Set_Atten(DWORD CardNum, DWORD ChanNum, DWORD Atten_Value)     // Up to 2 Channels Only
{
    if (!BrdNumOK(CardNum)) return;
//    P9050_HANDLE da11000_hPlx = GetBrdHandle(CardNum);

    Atten_Value = Atten_Value & 0x3F;  // Need only 6-bits

// Set ALL LE Low to Allow Serial Data + Set SER_DATA/CLOCK Low
    WritePort( 0,(WGR[0] & 0xFFF0));

// Clock in Data Starting with MSB First
    for (int x = 0; x < 6; x++) {
        if ((Atten_Value & 0x20) > 0) WritePort( 0,(WGR[0] | 0x1));  // Set SER_DATA to 1
        else  WritePort( 0,(WGR[0] & 0xFFFE)); // Set SER_DATA to 0

        WritePort( 0,(WGR[0] | 0x2)); WritePort( 0,(WGR[0] & 0xFFFD));  // Clock _|^|_
        Atten_Value = Atten_Value << 1;
    }

// Create Latch Pulse _|^|_
    if (ChanNum == 0x01) { WritePort( 0,(WGR[0] | 0x4 )); WritePort( 0,(WGR[0] & 0xFFFB)); }
    if (ChanNum == 0x02) { WritePort( 0,(WGR[0] | 0x8 )); WritePort( 0,(WGR[0] & 0xFFF7)); }
//    if (ChanNum == 0x04) { WritePort( 0,(WGR[0] | 0x10)); WritePort( 0,(WGR[0] & 0xFFEF)); }
//    if (ChanNum == 0x08) { WritePort( 0,(WGR[0] | 0x20)); WritePort( 0,(WGR[0] & 0xFFDF)); }
}

//---------------------------------------------------------------------------
//   Important note about DLL memory management when your DLL uses the
//   static version of the RunTime Library:
//
//   If your DLL exports any functions that pass String objects (or structs/
//   classes containing nested Strings) as parameter or function results,
//   you will need to add the library MEMMGR.LIB to both the DLL project and
//   any other projects that use the DLL.  You will also need to use MEMMGR.LIB
//   if any other projects which use the DLL will be perfomring new or delete
//   operations on any non-TObject-derived classes which are exported from the
//   DLL. Adding MEMMGR.LIB to your project will change the DLL and its calling
//   EXE's to use the BORLNDMM.DLL as their memory manager.  In these cases,
//   the file BORLNDMM.DLL should be deployed along with your DLL.
//
//   To avoid using BORLNDMM.DLL, pass string information using "char *" or
//   ShortString parameters.
//
//   If your DLL uses the dynamic version of the RTL, you do not need to
//   explicitly add MEMMGR.LIB as this will be done implicitly for you




